#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "event_loop.h"

void event_loop_conn_io_res(event_loop *el, event_loop_conn *conn) {
  do {
    ssize_t res = 0;
    size_t remain = conn->wbuf_size - conn->wbuf_sent;

    do {
      res = write(conn->fd, &conn->wbuf[conn->wbuf_sent], remain);
    } while (res < 0 && errno == EINTR);

    // EAGAIN doesn't mean there was an error, but rather that there's no more
    // data right now, but there might be more later, aka "try again later"
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // If it's not EGAIN, there was an error writing so we simply end the
    // request
    if (res < 0) {
      conn->state = event_loop_conn_state_end;
      return;
    }

    conn->wbuf_sent += (size_t)res;

    // After writing the entire buffer, we run the write_data command to receive
    // new data, or exit the loop
    if (conn->wbuf_sent == conn->wbuf_size) {
      conn->wbuf_sent = 0;
      conn->wbuf_size = 0;

      el->write_data(conn);
    }
  } while (conn->state == event_loop_conn_state_res);
}

/**
 * Read new data into the read buffer. This command performs at most one
 * successful read syscall.
 *
 * Returns whether the function should be retried immediately or not.
 */
void event_loop_conn_io_req(event_loop *el, event_loop_conn *conn) {
  do {
    // Move remaining data to start of buffer
    memmove(conn->rbuf, &conn->rbuf[conn->rbuf_read],
            conn->rbuf_size - conn->rbuf_read);
    conn->rbuf_size -= conn->rbuf_read;
    conn->rbuf_read = 0;

    ssize_t res;
    size_t cap = EVENT_LOOP_BUFFER_SIZE - conn->rbuf_size;

    // Try to read at most cap bytes from the file descriptor
    do {
      res = read(conn->fd, &conn->rbuf[conn->rbuf_size], cap);
    } while (res < 0 && errno == EINTR);

    // EGAIN means we try again later
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // Any other negative error message means the read errored out. If res is 0,
    // we've reached the end of the input which (usually) means the remote peer
    // has closed the connection. Either way, we close the connection.
    if (res <= 0) {
      conn->state = event_loop_conn_state_end;

      return;
    }

    conn->rbuf_size += (size_t)res;

    // This loop allows processing multiple requests from a single read buffer
    while (el->handle_data(conn))
      ;
  }
  // We can keep reading as long as we're in request mode
  while (conn->state == event_loop_conn_state_req);
}

void event_loop_conn_io(event_loop *el, event_loop_conn *conn) {
  switch (conn->state) {
  case event_loop_conn_state_req:
    event_loop_conn_io_req(el, conn);
    break;
  case event_loop_conn_state_res:
    event_loop_conn_io_res(el, conn);
    break;
  case event_loop_conn_state_end:;
  }
}
