#include <sys/stat.h>
#include <sys/types.h>

#include "crow.h"

extern "C" {
#include "trie.h"
}

static const std::string file_path = "lander.data";
static const std::string index_page = R"(
<!DOCTYPE html>
<html>
  <body>
    <h1>r8r.be</h1>
    <p>This is the URL shortener and pastebin accompanying my site, <a href="https://rustybever.be">The Rusty Bever</a>.</p>
  </body>
</html>
)";

#define ENV(var, env_var)                                                      \
  const char *_##var = getenv(env_var);                                        \
  if (_##var == NULL) {                                                        \
    printf("Missing environment variable %s.\n", env_var);                     \
    return 1;                                                                  \
  }                                                                            \
  const std::string var = std::string(_##var);

#define AUTH()                                                                 \
  std::string provided_api_key = req.get_header_value("X-Api-Key");            \
  if (api_key.compare(provided_api_key) != 0) {                                \
    return crow::response(crow::status::UNAUTHORIZED);                         \
  }

crow::response add_redirect(std::string base_url, Trie *trie, const char *url,
                            bool secure) {
  Entry *new_entry = entry_new(Redirect, url);

  // The key already gets copied into the trie, so this pointer is safe to use
  // ever after unlocking the trie
  trie_wlock(trie);
  char *key;
  TrieExitCode res = trie_add_random(trie, &key, new_entry, secure);
  trie_unlock(trie);

  if (res != Ok) {
    return crow::response(crow::status::INTERNAL_SERVER_ERROR);
  }

  std::string out = base_url + key;
  free(key);

  return crow::response(out);
}

bool store_paste(const char *key, const char *body) {
  // Write paste contents to file
  std::fstream file;
  file.open(std::string("pastes/") + key, std::ios_base::out);

  if (!file.is_open()) {
    return false;
  }

  file << body;
  file.close();

  return true;
}

crow::response add_paste(std::string base_url, Trie *trie, const char *body,
                         bool secure) {
  Entry *new_entry = entry_new(Paste, "");

  trie_wlock(trie);
  char *key;
  TrieExitCode res = trie_add_random(trie, &key, new_entry, secure);
  trie_unlock(trie);

  if (res != Ok) {
    return crow::response(crow::status::INTERNAL_SERVER_ERROR);
  }

  if (!store_paste(key, body)) {
    return crow::response(crow::status::INTERNAL_SERVER_ERROR);
  }

  std::string out = base_url + key;
  free(key);

  return crow::response(out);
}

int main() {
  // Initialize random seed for generating URLs
  srand(time(NULL));

  ENV(api_key, "LANDER_API_KEY");
  ENV(base_url, "LANDER_BASE_URL");

  std::cout << "Initializing trie from file '" << file_path << "'..."
            << std::endl;

  // Initialize trie and populate from data file
  Trie *trie;
  int res = trie_init(&trie, file_path.c_str());

  if (res != 0) {
    std::cout << "An error occured while initializing the trie." << std::endl;

    exit(1);
  }

  std::cout << "Added " << trie_size(trie) << " entries to trie." << std::endl;

  // Create pastes directory if not present
  // TODO don't just ignore errors here
  mkdir("pastes", 0700);

  crow::SimpleApp app;
  app.loglevel(crow::LogLevel::Info);

  CROW_ROUTE(app, "/").methods(crow::HTTPMethod::Get)(
      []() { return crow::response("html", index_page); });

  // Serve an entry
  CROW_ROUTE(app, "/<string>")
      .methods(crow::HTTPMethod::Get)(
          [trie](crow::response &res, std::string key) {
            trie_rlock(trie);
            Entry *entry;
            TrieExitCode status = trie_search(trie, &entry, key.c_str());

            if (status == Ok) {
              if (entry->type == Redirect) {
                res.redirect(entry->string);
              } else if (entry->type == Paste) {
                res.set_static_file_info("pastes/" + key);
              }
            } else {
              res.code = 404;
            }

            res.end();
            trie_unlock(trie);
          });

  // Add a new Redirect with a short randomly generated key
  CROW_ROUTE(app, "/s/")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request req) {
            AUTH();

            return add_redirect(base_url, trie, req.body.c_str(), false);
          });

  // Add a new Redirect with a long randomly generated key
  CROW_ROUTE(app, "/sl/")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request req) {
            AUTH();

            return add_redirect(base_url, trie, req.body.c_str(), true);
          });

  // Add a new Redirect with a given key
  CROW_ROUTE(app, "/s/<string>")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request &req, std::string key) {
            AUTH();

            Entry *new_entry = entry_new(Redirect, req.body.c_str());

            trie_wlock(trie);
            TrieExitCode status = trie_add(trie, key.c_str(), new_entry);
            trie_unlock(trie);

            switch (status) {
            case Ok:
              return crow::response(base_url + key);
            case AlreadyPresent:
              return crow::response(crow::status::CONFLICT);
            default:
              return crow::response(crow::status::INTERNAL_SERVER_ERROR);
            }
          });

  // Add a new Paste with a short randomly generated key
  CROW_ROUTE(app, "/p/")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request &req) {
            AUTH();

            return add_paste(base_url, trie, req.body.c_str(), false);
          });

  // Add a new Paste with a long randomly generated key
  CROW_ROUTE(app, "/pl/")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request &req) {
            AUTH();

            return add_paste(base_url, trie, req.body.c_str(), true);
          });

  // Add a paste with a given key
  CROW_ROUTE(app, "/p/<string>")
      .methods(crow::HTTPMethod::Post)(
          [api_key, base_url, trie](const crow::request &req, std::string key) {
            AUTH();

            Entry *new_entry = entry_new(Paste, "");
            trie_wlock(trie);
            TrieExitCode status = trie_add(trie, key.c_str(), new_entry);
            trie_unlock(trie);

            if (status != Ok) {
              return crow::response(crow::status::CONFLICT);
            }

            if (!store_paste(key.c_str(), req.body.c_str())) {
              return crow::response(crow::status::INTERNAL_SERVER_ERROR);
            }

            return crow::response(base_url + key);
          });
  app.port(18080).multithreaded().run();
}
